/*RELAB 0D = RElativistic LAttice Boltzmann solver in 0+1 spacetime D

by

Paul Romatschke

version 0.31 December 2011 (still beta!)

This code is a relativistic fluid dynamics solver based on the 
Boltzmann equation, derived in

P. Romatschke, M. Mendoza, S. Succi:
'A fully relativistic lattice Boltzmann algorithm'
Published in Phys.Rev. C84 (2011) 034903
e-Print: arXiv:1106.1093 [nucl-th]

If you use this code or some part of it, please be so kind as
to add a reference to this article. Thank you.


Copyright of this code is granted provided you keep this disclaimer.


*/

#include <relab0D.h>


fstream Tout;

double T0;

// allocate memory
void allocateMemory() 
{

  printf("Info: ===> Allocating memory\n");
  
  amn = new double[Np*Nxi];
  fcurr = new double[Np*Nxi];
  fprime = new double[Np*Nxi];

  p = new double[Np];
  x = new double[Nxi];
  w_p = new double[Np];
  w_x = new double[Nxi];

  Lm1p = new double[Np*Np];
  Plx  = new double[Nxi*Nxi];
}

void cleanup()
{

  delete [] amn;
  delete [] fcurr;
  delete [] fprime;
  delete [] p;
  delete [] x;
  delete [] w_p;
  delete [] w_x;

  delete [] Lm1p;
  delete [] Plx;

}


//initialize momentum grid
void getpgrid()
{
  switch(Np)
    {
    case 1:
      printf("Info: Np=%i grid\n",Np);
      p[0]=4;
      break; 
    case 2:
      printf("Info: Np=%i grid\n",Np);
      p[0]=2.76393202250021;
      p[1]=7.23606797749979;
      break; 
    case 3:
      printf("Info: Np=%i grid\n",Np);
      p[0]=2.141216276717724;
      p[1]=5.315517126176786;
      p[2]=10.54326659710549;
      break;
    case 5:
      printf("Info: Np=%i grid\n",Np);
      p[0]=1.490554945186829;
      p[1]=3.581333812903247;
      p[2]=6.626996296823595;
      p[3]=10.944418003467705;
      p[4]=17.35669694161863;
    default:
      printf("Np=%i :Case not defined\n",Np);
    }

  for (int i=0;i<Np;i++)
    {
      double temp=(Np+1)*gsl_sf_laguerre_n (Np+1,3,p[i]);
      temp*=temp;
      w_p[i]=p[i]/temp*(Np+3)*(Np+2)*(Np+1);
    }

  switch (Nxi)
    {
    case 3:
      printf("Info: Nxi=%i grid\n",Nxi);
      x[0]=-0.7745966692414834;
      x[1]=0;
      x[2]=0.7745966692414834;
      break;
    case 4:
      printf("Info: Nxi=%i grid\n",Nxi);
      x[0]=-0.8611363115940526;
      x[1]=-0.33998104358485626;
      x[2]=0.33998104358485626;
      x[3]=0.8611363115940526;
      break;
    case 5:
      printf("Info: Nxi=%i grid\n",Nxi);
      x[0]=-0.9061798459386641;
      x[1]=-0.538469310105683;
      x[2]= 0;
      x[3]=0.538469310105683;
      x[4]=0.9061798459386641;
      break;
    case 6:
      printf("Info: Nxi=%i grid\n",Nxi);
      x[0]=-0.9324695142031519;
      x[1]=-0.6612093864662646;
      x[2]=-0.23861918608319693;
      x[3]=0.2386191860831969;
      x[4]=0.6612093864662646;
      x[5]=0.9324695142031519;
      break;
    default:
      printf("Nxi=%i : Case not defined\n",Nxi);
    }

  for (int i=0;i<Nxi;i++)
    {
      double temp=dPdx(Nxi,x[i]);
      temp*=temp*(1-x[i]*x[i]);
      w_x[i]=2./temp;
    }
   
}

double feq(double ptot, double T)
{
  double temp=exp(-ptot/T);
  return temp;
}



//initial conditions

void setInitCond(double T)
{
  printf("Info: ===> Setting initial conditions\n"); 


  for (int i=0;i<Np;i++)
    for (int j=0;j<Nxi;j++)
      {
	set_f(fcurr,i,j,feq(p[i]*theta,theta)*pow(T/theta,4.));
	set_f(fprime,i,j,feq(p[i]*theta,theta)*pow(T/theta,4.));
      }
      
}


//initialize array of Laguerre and Legendre Polynomials
void getLandParray()
{
  for(int m=0;m<Np;m++)
    for (int i=0;i<Np;i++)\
      set_sqrA(Np,Lm1p,m,i,gsl_sf_laguerre_n (m,3,p[i]));
  
  for(int l=0;l<Nxi;l++)
    for (int i=0;i<Nxi;i++)\
      set_sqrA(Nxi,Plx,l,i,gsl_sf_legendre_Pl (l,x[i]));

}

//get amns
double get_amn(double *f,int m,int l)
{
  double temp=0;
  
  double mult=(m+1)*(m+2)*(m+3.);
  for (int i=0;i<Np;i++)
    {
      double mex=exp(p[i]);
      for (int j=0;j<Nxi;j++)
	{       
	  temp+=(l+0.5)/mult*w_p[i]*w_x[j]*mex*get_sqrA(Np,Lm1p,m,i)*get_sqrA(Nxi,Plx,l,j)*get_f(f,i,j);
	}
    }
  return temp;
}

//S00 coefficient
double S00(double *f)
{
  double temp=0;

  
  for (int i=0;i<Np;i++)
    {
      double mex=exp(p[i]);
      for (int j=0;j<Nxi;j++)
	{       
	  temp+=w_p[i]*w_x[j]*mex*get_f(f,i,j)*(1+x[j]*x[j]);
	}
    }
  return temp/12.;
}

//Setaeta coefficient
double Setaeta(double *f)
{
  double temp=0;
  
  for (int i=0;i<Np;i++)
    {
      double mex=exp(p[i]);
      for (int j=0;j<Nxi;j++)
	{       
	  temp-=w_p[i]*w_x[j]*mex*get_f(f,i,j)*(3*pow(x[j],4)-8*x[j]*x[j]+1)/3.;
	}
    }
  return temp*15./24.;
}


//get amns
void set_amns(double *f)
{
  double temp=0;
  
  for (int m=0;m<Np;m++)
    for (int l=0;l<Nxi;l++)
      set_f(amn,m,l,get_amn(f,m,l));

}

double getT(double *f)
{
  double temp=0;
  double a00=get_amn(f,0,0);
   
  //printf("debug a00=%f\n",a00);

  temp=pow((a00),0.25);
  temp*=theta;
  return temp;
}

double Thydro(double t, double T0, double tau0, double etaos)
{
  double temp=pow(tau0/tau,1./3.);
  return T0*temp*(1+2./3.*etaos/(tau0*T0)*(1-temp*temp));
}

double tauR(double T,double etaos)
{
  double temp=5*etaos/T;
  return temp;
}


void dotheludwig(long int STEPS)
{

  for (long int cnt=0;cnt<STEPS;cnt++)
    {
      double Tcur;

      

      tau+=dt;
      
      double addnew;
      

      for (int i=0;i<Np;i++)
	for (int j=0;j<Nxi;j++)
	  {
	    double temp=exp(-p[i]);

	    addnew=S00(fprime)+Setaeta(fprime)*get_sqrA(Nxi,Plx,2,j);
	    addnew*=dt/tau;
	    set_f(fcurr,i,j,get_f(fprime,i,j)-addnew*temp);
	  }

      Tcur=getT(fcurr);
      
      Tout << tau/fmtoGeVinv << "\t";
      Tout << Tcur <<"\t";
      Tout << Tcur/Thydro(tau,T0,1.*fmtoGeVinv,0.) << "\n";

      if (cnt%100==99)
	printf("tau=%f Tcur=%f Thyd=%f Tid=%f\n",tau/fmtoGeVinv,Tcur,Thydro(tau,T0,1.*fmtoGeVinv,ETAOS),Thydro(tau,T0,1.*fmtoGeVinv,0.));

	//printf("tauR=%f\n",tauR(Tcur,ETAOS));

      for (int i=0;i<Np;i++)
	for (int j=0;j<Nxi;j++)
	  {
	    //double nfeq=feq(p[i]*theta,Tcur);
	    double nfeq=feq(p[i]*theta,theta)*pow(Tcur/theta,4.);
	    double omega=dt/tauR(Tcur,ETAOS);
	    set_f(fprime,i,j,get_f(fcurr,i,j)*(1-omega)+omega*nfeq);
	  }
      
    }
}



int main()
{

  //all units in GeV
  tau=1.0*fmtoGeVinv;
  dt=0.01*fmtoGeVinv;

  ETAOS=0.5;

  Np=1;
  Nxi=5;

  theta=0.5;

  allocateMemory();
  
  //open files

  Tout.open("Tout.dat",ios::out);
  Tout << "# time [fm/c]" << "\t";
  Tout << "# T_sim [GeV]" << "\t";
  Tout << "# T/T_id_hydro" << "\n";

  getpgrid();
  getLandParray();

  
  //print_grid();
  T0=0.32971615;

  setInitCond(T0);

  //print_f(fcurr);
  
  printf("Info: ===> Starting\n"); 

  double Tcur=getT(fcurr);
      
  printf("tau=%f Tcur=%f Thyd=%f Tid=%f\n",tau/fmtoGeVinv,Tcur,Thydro(tau,T0,1.*fmtoGeVinv,ETAOS),Thydro(tau,T0,1.*fmtoGeVinv,0.));

  dotheludwig(10000);

  Tout.close();

  cleanup();
  
  printf("Info: ===> Done\n"); 

  return 0;
}



/********************************
CHANGELOG

25.01.2011 Started Coding; finished 0+1D boost-invariant algorithm (v0.1)
21.02.2011 Revised algorithm: used L^(3) instead of L^(1) and 
used explicit ansatz for f_eq rather than full form (v0.2)
22.02.2011 Revised algorithm: uses S00, Setaeta instead of amn coefficients
(v0.3)
08.12.2011 Some bug fixes: now runs for arbitrary T0 (v0.31)
08.12.2011 WARNING: there seems to be a bug in the free-streaming part (NOT FIXED)
*********************************/
